import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from keras import callbacks
from keras.models import Sequential,Model
from keras.layers import LSTM,Input
from keras.layers import Dense,Lambda,Flatten,GaussianNoise,Reshape,TimeDistributed
import keras.backend as K
import sys
from scipy.stats import multivariate_normal
from scipy.optimize import minimize
from sklearn.metrics import mean_squared_error
import functools
from keras.models import load_model
from pathlib import Path


size_limit=(32,32)
V_range=(-10,10)
T=10
Hz = 3
dt = 1.0/Hz
c = 0.01 # Resistance Coefficient
m = int(T/dt) # number of measurements

noise_p_train=1
outlier_p_train=0.05
spn_train=0.2
train_seq_len=10000

noise_p_test=1
outlier_p_test=0.15
spn_test=0.4
test_seq_len=3000

def dataGenerator():
    vx,vy=np.random.uniform(V_range[0],V_range[1],2)
    direction=np.sign([vx,vy])
    px,py=np.random.uniform(0,size_limit[0],2)
    Xr=[]
    Yr=[]
    Vx=[]
    Vy=[]
    for i in range(int(m)):
        if px>size_limit[0] :
            direction[0]=-1*direction[0]
            #vx=-vx
            px=size_limit[0]-0.02
        if px<0.01:
            direction[0]=-1*direction[0]
            #vx=-vx
            px=0.02
        if py >=size_limit[1]:
            #vy=-vy
            direction[1]=-1*direction[1]
            py=size_limit[1]-0.02
        if py <0:
            #vy=-vy
            direction[1]=-1*direction[1]
            py=0.02    
        Xr.append(px)
        Yr.append(py)
        Vx.append(vx)
        Vy.append(vy)

        vx =direction[0]*(abs(vx)-c*(vx**2))
        vy =direction[1]*(abs(vy)-c*(vy**2))
        px += vx*dt
        py += vy*dt
  
    return Xr,Yr,Vx,Vy

def addNoise(Xr,Yr,noise_p,spn):
    Xm=np.copy(Xr)
    Ym=np.copy(Yr)

    p=np.random.rand(m)
    Xm[p<noise_p]+=spn * (np.random.randn(np.sum(p<noise_p)))
    Ym[p<noise_p]+=spn * (np.random.randn(np.sum(p<noise_p)))
    return Xm,Ym

def addOutlier(Xr,Yr,outlier_p):
    Xm=np.copy(Xr)
    Ym=np.copy(Yr)

    p=np.random.rand(m)
    Xm[p<outlier_p] = np.random.rand(np.sum(p<outlier_p))*size_limit[0]
    Ym[p<outlier_p] = np.random.rand(np.sum(p<outlier_p))*size_limit[1]
    return Xm,Ym

def datasetGenerator(seq_len,noise_p,outlier_p,spn):
    clearDataset=[]
    NoisyDataset=[]
    NoisyOutlierDataset=[]
    for i in range(seq_len):
        Xr,Yr,Vx,Vy=dataGenerator()
        Xn,Yn=addNoise(Xr,Yr,noise_p,spn)
        Xo,Yo=addOutlier(Xn,Yn,outlier_p)

        Xn=np.reshape(Xn,(-1,1))
        Yn=np.reshape(Yn,(-1,1))

        Xo=np.reshape(Xo,(-1,1))
        Yo=np.reshape(Yo,(-1,1))

        Xr=np.reshape(Xr,(-1,1))
        Yr=np.reshape(Yr,(-1,1))
        Vx=np.reshape(Vx,(-1,1))
        Vy=np.reshape(Vy,(-1,1))

        NoisyDataset.append(np.concatenate((Xn,Yn,Vx,Vy),axis=-1))
        NoisyOutlierDataset.append(np.concatenate((Xo,Yo,Vx,Vy),axis=-1))
        clearDataset.append(np.concatenate((Xr,Yr,Vx,Vy),axis=-1))
    return np.array(clearDataset),np.array(NoisyDataset),np.array(NoisyOutlierDataset)

clean_train,Noisy_train,NoisyOutlier_train=datasetGenerator(train_seq_len,noise_p_train,outlier_p_train,spn_train)
clean_test,_,NoisyOutlier_test=datasetGenerator(test_seq_len,noise_p_test,outlier_p_test,spn_test)

sa_train=np.std((clean_train[:,1:,2:]-clean_train[:,:-1,2:])/dt)
print(sa_train)
sa_test=np.std((NoisyOutlier_test[:,1:,2:]-NoisyOutlier_test[:,:-1,2:])/dt)
print(sa_test)

dt = (1.0/Hz)
G = np.matrix([[0.5*dt**2],
               [0.5*dt**2],
               [dt],
               [dt]])

Q = G*G.T*sa_train**2
y = multivariate_normal(mean=np.array([0,0,0,0]), cov=Q,allow_singular=True)
y.rvs()

X_train=[clean_train[:,:,:2],Noisy_train[:,:,:2],NoisyOutlier_train[:,:,:2]]
Y_train=[clean_train[:,:,:2],clean_train[:,:,:2],clean_train[:,:,:2]]
X_test=NoisyOutlier_test[:,:,:2]
Y_test=clean_test[:,:,:2]
names=["Clean","Noisy","NoisyWithOutlier"]

epochs=50
iterations=10
batch_size=32
outputs_dim=2
input_dim=2
m=30

def root_mean_squared_error(y_true, y_pred):
    return K.sqrt(K.mean(K.square(y_pred - y_true))) 

Path("Data").mkdir(parents=True, exist_ok=True)
np.save("Data/X_train.npy",X_train)
np.save("Data/Y_train.npy",Y_train)
np.save("Data/X_test.npy",X_test)
np.save("Data/Y_test.npy",Y_test)

X_train=np.load("Data/X_train.npy")
Y_train=np.load("Data/Y_train.npy")
X_test=np.load("Data/X_test.npy")
Y_test=np.load("Data/Y_test.npy")
names=["Clean","Noisy","NoisyWithOutlier"]

savePath=r"Inverse Tracking Results/DM_TimeDistributed/"
Path(savePath).mkdir(parents=True, exist_ok=True)

for i in range(len(X_train)):
#for i in range(1):
    X=X_train[i]
    Y=Y_train[i]
    name=names[i]
    PRED=[]
    print(name)
    for j in range(iterations):
        print(i,j)
        model = Sequential()
        model.add(LSTM(64, activation='relu', input_shape=(m, input_dim),return_sequences=True))
        model.add(TimeDistributed(Dense(32)))
        model.add(TimeDistributed(Dense(16)))
        model.add(TimeDistributed(Dense(2)))
        model.compile(optimizer='adam', loss='mse',metrics=[root_mean_squared_error])
        model.summary()
        model.fit(X, Y, epochs=epochs, verbose=0, validation_split=0.2)
        #model.save("%s%s_DM_%d_%d.h5" % (TenTimes_weights_path,model_name,n_step,num))
        PRED.append(model.predict(X_test))
        print(root_mean_squared_error(Y_test,PRED[-1]))
    np.save('%s%s' % (savePath,name),np.array(PRED))

savePath=r"Inverse Tracking Results/DC_TimeDistributed/"
Path(savePath).mkdir(parents=True, exist_ok=True)

for i in range(1,len(X_train)):
    X=X_train[i]
    Y=Y_train[i]
    name=names[i]
    PRED=[]
    for j in range(iterations):
        print(i,j)
        inp = Input(shape=(30, input_dim))
        x=LSTM(64, activation='relu', input_shape=(m, input_dim),return_sequences=True)(inp)
        #x=LSTM(100, activation='relu',return_sequences=True)(x)
        #x=Flatten()(x)
        x=TimeDistributed(Dense(32))(x)
        x=TimeDistributed(Dense(16))(x)
        x=TimeDistributed(Dense(2))(x)
        #x=Reshape((-1,outputs_dim))(x)
        #f=Lambda(forward)(x)
        if i!=0:
            f=TimeDistributed(GaussianNoise(0.2))(x)
        else:
            f=TimeDistributed(Lambda(lambda x1: x1))(x)
        model=Model(inp,f)
        model.compile(optimizer='adam', loss='mse',metrics=['acc'])
        model.summary()
        model.fit(X, X, epochs=epochs, verbose=0, validation_split=0.2)
        model_P=Model(inp,x)
        PRED.append(model_P.predict(X_test))
        print(root_mean_squared_error(Y_test,PRED[-1]))
    np.save('%s%s' % (savePath,name),np.array(PRED))



sample_num=4000
box=[-16,48,-16,48]
boxSize=(32,32)
data=(np.random.rand(sample_num,2)*np.array([box[1]-box[0],box[3]-box[2]]))+np.array([box[0],box[2]])
#np.save(save_data_path+"reg_data.npy",data)
c=functools.reduce(np.logical_or,((data[:,0]>=32),(data[:,0]<=0),(data[:,1]>=32),(data[:,1]<=0)))
plt.figure(figsize=(6,6),dpi=100)
plt.scatter(data[np.argwhere(c==0),0],data[np.argwhere(c==0),1],s=4,label="In the box",c='black')
plt.scatter(data[np.argwhere(c==1),0],data[np.argwhere(c==1),1],s=4,label="Out of the box",c='lightgrey')
plt.legend()
plt.show()
print(np.bincount(c))
print(data.shape,c.shape)


Reg_model = Sequential()
Reg_model.add(Dense(8,activation='relu',input_dim=2))
Reg_model.add(Dense(4,activation='relu'))
Reg_model.add(Dense(1,activation='sigmoid'))
Reg_model.compile(optimizer='adam', loss='binary_crossentropy',metrics=['acc'])
Reg_model.summary()

Reg_model.fit(data, c, epochs=200, verbose=1,validation_split=0.2)
m=Reg_model.evaluate(data,c)
print(m)
Reg_model.save("regularizer.h5")


regularizer_model=load_model("regularizer.h5")


savePath=r"Inverse Tracking Results/DR-Genetic-Algorithm/"
Path(savePath).mkdir(parents=True, exist_ok=True)

def cal_pop_fitness(m, pop):
    m_hat=pop+(0.2 * (np.random.randn(pop.shape[0],pop.shape[1],outputs_dim)))
    s=np.mean((m_hat-m)**2,axis=1)
    reg=np.mean(regularizer_model.predict(pop.reshape(-1,2)).reshape(-1,30),axis=1)
    fitness=np.mean(s,axis=1)+reg
    return np.array(fitness).reshape((-1,))
    
def select_mating_pool(pop, fitness, num_parents):
    idx=np.argsort(fitness)
    return pop[idx[:num_parents]]

def crossover(parents,offspring_size):
    offspring = np.empty(offspring_size)
    crossover_point = np.uint8(offspring_size[1]/2)
    parent1_idx=np.random.randint(parents.shape[0],size=offspring_size[0])
    parent2_idx=np.random.randint(parents.shape[0],size=offspring_size[0])
    offspring[:, :crossover_point] = parents[parent1_idx, :crossover_point]
    offspring[:, crossover_point:] = parents[parent2_idx, crossover_point:]
    return offspring

def mutation(offspring_crossover,Pm=0.3,box_size=[32,32]):
    mutation_p=np.random.rand(offspring_crossover.shape[0])
    mutation_idx=np.argwhere(mutation_p<Pm).reshape((-1,))
    offspring_crossover[mutation_idx]= np.random.rand(mutation_idx.shape[0],30,2)*box_size
    return offspring_crossover


res=np.zeros((10,Y_test.shape[0],30,outputs_dim))
for i in range(10):
    for j in range(X_test.shape[0]):
        pop_size = 50
        num_parents_mating = 30
        feature_size=2
        box_size=[32,32]
        num_generations = 100
        new_population = np.random.rand(pop_size,30,2)+X_test[j]
        for generation in range(num_generations):
            fitness = cal_pop_fitness(X_test[j], new_population)
            parents = select_mating_pool(new_population, fitness, num_parents_mating)
            offspring_crossover = crossover(parents,offspring_size=(pop_size-parents.shape[0],30,feature_size))
            offspring_mutation = mutation(offspring_crossover)
            new_population=np.concatenate([parents,offspring_mutation])
            fitness = cal_pop_fitness(X_test[j], new_population)
        print(i,j,np.min(fitness))
        res[i,j]=new_population[np.argmin(fitness)]
np.save('%s%s' % (savePath,"res.npy"),np.array(res))

regularizer_model=load_model("regularizer.h5")

savePath=r"Inverse Tracking Results/DR-Nelder-Mead/"
Path(savePath).mkdir(parents=True, exist_ok=True)

a=[50,50]
regularizer_model.predict(np.reshape(a,(1,-1)))

def func(p,m):
    p=p.reshape((-1,2))
    reg_term=regularizer_model.predict(p)
    m_hat=p+(0.2 * (np.random.randn(30,outputs_dim)))
    s=np.mean((m-m_hat)**2+reg_term)
    return s

res=np.zeros((Y_test.shape[0],30,outputs_dim))
print(res.shape)
for j in range(X_test.shape[0]):
    x0=X_test[j]
    print(j)
    Res = minimize(func, x0, method='Nelder-Mead', options={'maxiter':100}, args=(Y_test[j]))
    res[j]=Res.x.reshape((-1,2))
    print(np.mean((res[j]-Y_test[j])**2))

np.save('%s%s' % (savePath,"res.npy"),np.array(res))

results={
    "DC Clean Dist":"Inverse Tracking Results/DC_TimeDistributed/Clean.npy",
    "DC Noisy Dist":"Inverse Tracking Results/DC_TimeDistributed/Noisy.npy",
    "DC Noisy Outlier Dist":"Inverse Tracking Results/DC_TimeDistributed/NoisyWithOutlier.npy",
    "DM Clean Dist":"Inverse Tracking Results/DM_TimeDistributed/Clean.npy",
    "DM Noisy Dist":"Inverse Tracking Results/DM_TimeDistributed/Noisy.npy",
    "DM Noisy Outlier Dist":"Inverse Tracking Results/DM_TimeDistributed/NoisyWithOutlier.npy",
    "DR_NM Noisy Outlier":"Inverse Tracking Results/DR-Nelder-Mead/res.npy",
    "DR_GA Noisy Outlier":"Inverse Tracking Results/DR-Genetic-Algorithm/res_9.npy"
}

for key, value in results.items():
    print(key)
    pred=np.load(value)
    if "GA" in key:
        rmse=np.mean(np.mean(np.mean(np.sqrt((pred[:,:1000]-Y_test[:1000])**2),axis=-1),axis=-1),axis=-1)
    else:
        rmse=np.mean(np.mean(np.mean(np.sqrt((pred-Y_test)**2),axis=-1),axis=-1),axis=-1)
    print(rmse)
    Path("/Inverse Tracking Results/RMSE/").mkdir(parents=True, exist_ok=True)
    np.save("/Inverse Tracking Results/RMSE/"+key,rmse)
